import logging
import random
from collections import deque, namedtuple

from slvcodec import cocotb_wrapper as cocotb
from slvcodec import cocotb_helper

from axilent import comms


logger = logging.getLogger(__name__)


def signal_to_integer(s):
    if not s.value.is_resolvable:
        return None
    else:
        return s.value.integer


def make_axi_signals_from_prefixs(dut, m2s_prefix, s2m_prefix):
    axi_signals = {
        'awvalid': getattr(dut, m2s_prefix + 'awvalid'),
        'awaddr': getattr(dut, m2s_prefix + 'awaddr'),
        'awready': getattr(dut, s2m_prefix + 'awready'),
        'wvalid': getattr(dut, m2s_prefix + 'wvalid'),
        'wdata': getattr(dut, m2s_prefix + 'wdata'),
        'wready': getattr(dut, s2m_prefix + 'wready'),
        'bvalid': getattr(dut, s2m_prefix + 'bvalid'),
        'bresp': getattr(dut, s2m_prefix + 'bresp'),
        'bready': getattr(dut, m2s_prefix + 'bready'),
        'arvalid': getattr(dut, m2s_prefix + 'arvalid'),
        'araddr': getattr(dut, m2s_prefix + 'araddr'),
        'arready': getattr(dut, s2m_prefix + 'arready'),
        'rvalid': getattr(dut, s2m_prefix + 'rvalid'),
        'rresp': getattr(dut, s2m_prefix + 'rresp'),
        'rdata': getattr(dut, s2m_prefix + 'rdata'),
        'rready': getattr(dut, m2s_prefix + 'rready'),
        }
    return axi_signals


def make_axi_signals_from_interfaces(m2s, s2m):
    axi_signals = {
        'awvalid': m2s.awvalid,
        'awaddr': m2s.awaddr,
        'awready': s2m.awready,
        'wvalid': m2s.wvalid,
        'wdata': m2s.wdata,
        'wready': s2m.wready,
        'bvalid': s2m.bvalid,
        'bresp': s2m.bresp,
        'bready': m2s.bready,
        'arvalid': m2s.arvalid,
        'araddr': m2s.araddr,
        'arready': s2m.arready,
        'rvalid': s2m.rvalid,
        'rresp': s2m.rresp,
        'rdata': s2m.rdata,
        'rready': m2s.rready,
        }
    return axi_signals


QueueInfo = namedtuple('QueueInfo', ['throttle'])


class CocotbHandler(object):
    '''
    This handler receives `Command` objects and sends their AXI
    commands into a cocotb simulation.
    '''

    def __init__(self, clock_signal, axi_signals=None,
                 dut=None, m2s_prefix=None, s2m_prefix=None,
                 m2s=None, s2m=None, random_seed=None, helper=None):
        '''
        Args: 
          `clock_signal`: The signal for the clock.
          `axi_signals`: A dictionary relating signal names to the cocotb signal objects.
          `dut`, `m2s_prefix`, `s2m_prefix`: Alternately the axi signals can be constructed
             from the dut interface and the names of the prefixes for the m2s and s2m signals.
          `m2s`, `s2m`: Or the objects for the interfaces to the m2s and s2m signal bundles
             can be directly passed.
        '''
        if axi_signals is not None:
            assert m2s_prefix is None
            assert s2m_prefix is None
            assert m2s is None
            assert s2m is None
            self.signals = axi_signals
        elif m2s_prefix is not None:
            assert s2m_prefix is not None
            assert m2s is None
            assert s2m is None
            self.signals = make_axi_signals_from_prefixs(dut, m2s_prefix, s2m_prefix)
        else:
            assert m2s is not None
            assert s2m is not None
            assert s2m_prefix is None
            self.signals = make_axi_signals_from_interfaces(m2s, s2m)
        if helper is None:
            helper = cocotb_helper.TaskHelper()
        self.helper = helper
        self.clock_signal = clock_signal
        # Queues for storing 'w', 'aw', 'ar' commands until
        # they are sent into the simulation.
        self.w_queue = deque()
        self.aw_queue = deque()
        self.ar_queue = deque()
        # Queues for storing expectations of replies on the
        # 'b' and 'r' channels.
        self.b_queue = deque()
        self.r_queue = deque()
        self.command_queue = deque()
        self.random_seed = random_seed

    def start(self, aw_throttle=0.8, w_throttle=0.8, b_throttle=0.8,
              ar_throttle=0.8, r_throttle=0.8):
        """
        Activate the handler and start applying queued commands to the
        simulation.
        """
        self.helper.fork(self.process_queue_out(
            self.aw_queue,
            self.signals['awvalid'],
            self.signals['awready'],
            self.signals['awaddr'],
            QueueInfo(throttle=aw_throttle),
            random_seed=None if self.random_seed is None else self.random_seed + 10000,
        ))
        self.helper.fork(self.process_queue_out(
            self.w_queue,
            self.signals['wvalid'],
            self.signals['wready'],
            self.signals['wdata'],
            QueueInfo(throttle=w_throttle),
            random_seed=None if self.random_seed is None else self.random_seed + 20000,
        ))
        self.helper.fork(self.process_queue_in(
            self.b_queue,
            self.signals['bvalid'],
            self.signals['bready'],
            self.signals['bresp'],
            QueueInfo(throttle=b_throttle),
            random_seed=None if self.random_seed is None else self.random_seed + 30000,
        ))
        self.helper.fork(self.process_queue_out(
            self.ar_queue,
            self.signals['arvalid'],
            self.signals['arready'],
            self.signals['araddr'],
            QueueInfo(throttle=ar_throttle),
            random_seed=None if self.random_seed is None else self.random_seed + 40000,
        ))
        self.helper.fork(self.process_queue_in(
            self.r_queue,
            self.signals['rvalid'],
            self.signals['rready'],
            self.signals['rresp'],
            QueueInfo(throttle=r_throttle),
            self.signals['rdata'],
            random_seed=None if self.random_seed is None else self.random_seed + 50000,
        ))

    async def send(self, command):
        '''
        Sends a Command objects to the simulation and processes the responses.
        '''
        read_events = []
        write_events = []
        for ac in command.get_axi_commands():
            if isinstance(ac, comms.FakeWaitCommand):
                for dummy_index in range(ac.clock_cycles):
                    await self.helper.RisingEdge(self.clock_signal)
            else:
                assert(ac.readorwrite in (comms.WRITE_TYPE, comms.READ_TYPE))
                if ac.readorwrite == comms.WRITE_TYPE:
                    for offset, d in enumerate(ac.data):
                        address = ac.start_address
                        if not ac.constant_address:
                            if ac.address_by_word:
                                address += offset
                            else:
                                address += offset * 4
                        write_events.append(self.submit_write(address, d))
                else:
                    for index in range(ac.length):
                        address = ac.start_address
                        if not ac.constant_address:
                            if ac.address_by_word:
                                address += index
                            else:
                                address += index * 4
                        read_events.append(self.submit_read(address))

        for index, event in enumerate(write_events):
            await event.wait()
        for index, event in enumerate(read_events):
            await event.wait()
        write_event_results = [f.data for f in write_events]
        write_responses = deque(comms.AxiResponse(length=1, data=[None if data is None else int(data)], resp=resp)
                           for resp, data in write_event_results)

        read_event_results = [f.data for f in read_events]
        read_responses = deque(comms.AxiResponse(length=1, data=[None if data is None else int(data)], resp=resp)
                          for resp, data in read_event_results)
        command.process_responses(read_responses, write_responses)
        return command.future.result()

    def submit_write(self, address, value):
        event = self.helper.Event()
        self.aw_queue.append(address)
        self.w_queue.append(value)
        self.b_queue.append(event)
        return event

    async def write(self, address, value):
        event = self.submit_write(address, value)
        await event.wait()
        response, data = event.data
        if response != comms.OKAY:
            raise comms.AxiResponseException('Bad response: {}'.format(response))
        return data

    def submit_read(self, address):
        event = self.helper.Event()
        self.ar_queue.append(address)
        self.r_queue.append(event)
        return event

    async def read(self, address):
        event = self.submit_read(address)
        await event.wait()
        response, data = event.data
        if response != comms.OKAY:
            raise comms.AxiResponseException('Bad response: {}'.format(response))
        return data

    @cocotb_helper.killable
    async def process_queue_out(helper, self, queue, valid_signal, ready_signal, data_signal, queue_info,
                                random_seed=None):
        """
        Send data from a queue into the simulation using valid/ready hand shaking.
        """
        rnd = random if random_seed is None else random.Random(random_seed)
        while True:
            if queue:
                value = queue.popleft()
                while True:
                    valid = 1 if rnd.random() < queue_info.throttle else 0
                    valid_signal.value = valid
                    if valid:
                        data_signal.value = value
                    await helper.ReadOnly()
                    assert str(ready_signal.value) in ('0', '1')
                    consumed = (ready_signal.value == 1) and valid
                    await helper.RisingEdge(self.clock_signal)
                    if consumed:
                        break
            else:
                valid_signal.value = 0
                await helper.RisingEdge(self.clock_signal)

    @cocotb_helper.killable
    async def process_queue_in(helper, self, queue, valid_signal, ready_signal, resp_signal, queue_info,
                               data_signal=None, random_seed=None):
        """
        Take data from the simulation using valid/ready shaking.
        Take an 'Event' from a queue and apply the received data to that event.
        """
        rnd = random if random_seed is None else random.Random(random_seed)
        while True:
            ready = 1 if rnd.random() < queue_info.throttle else 0
            ready_signal.value = ready
            await helper.ReadOnly()
            assert str(valid_signal.value) in ('0', '1')
            if (valid_signal.value == 1) and ready:
                event = queue.popleft()
                if data_signal is not None:
                    event.set((resp_signal.value, data_signal.value))
                else:
                    event.set((resp_signal.value, None))
            await helper.RisingEdge(self.clock_signal)



import logging
import random

from slvcodec import fusesoc_wrapper
from slvcodec import cocotb_wrapper as cocotb

from axilent import comms


logger = logging.getLogger(__name__)


class AxiAdderComm(object):
    '''
    Class to communicate with the AxiAdder module.
    '''

    INTA_ADDRESS = 0
    INTB_ADDRESS = 1
    INTC_ADDRESS = 2

    def __init__(self, address_offset, handler):
        '''
        `address_offset` is any addition that is made to the address that is
        consumed during routing.
        `handler` is the object responsible for dispatching the commands.
        '''
        self.handler = handler
        self.address_offset = address_offset
        self.addresses = {
            'intA': address_offset + self.INTA_ADDRESS,
            'intB': address_offset + self.INTB_ADDRESS,
            'intC': address_offset + self.INTC_ADDRESS,
        }

    def add_numbers(self, a, b):
        '''
        A complex complex command that write to two registers and
        then reads from another.
        Sets 'a' and 'b' then reads 'c' (should be a+b)
        '''
        command = AddNumbersCommand(a, b, self.addresses)
        self.handler.send(command)
        return command.future

    @cocotb.coroutine
    async def add_numbers_async(self, a, b):
        write_a_and_b_command = comms.CombinedCommand((
            comms.SetUnsignedCommand(
                address=self.addresses['intA'], value=a,
                description='Setting A in AddNumbers',
            ),
            comms.SetUnsignedCommand(
                address=self.addresses['intB'], value=b,
                description='Setting B in AddNumbers',
            ),
            ), description='Setting A and B in AddNumbers')
        await self.handler.send(write_a_and_b_command)
        read_c_command = comms.GetUnsignedCommand(
            address=self.addresses['intC'],
            description='Getting C from AddNumbers',
        )
        result = await self.handler.send(read_c_command)
        return result


class AddNumbersCommand(comms.CombinedCommand):
    '''
    A command that writes to the intA and intB registers
    and then reads from the intC register.
    The effect is the add the two inputs.
    '''

    def __init__(self, a, b, addresses):
        write_a_command = comms.SetUnsignedCommand(
            address=addresses['intA'], value=a,
            description='Setting A in AddNumbers',
        )
        write_b_command = comms.SetUnsignedCommand(
            address=addresses['intB'], value=b,
            description='Setting B in AddNumbers',
        )
        read_c_command = comms.GetUnsignedCommand(
            address=addresses['intC'],
            description='Getting C from AddNumbers',
        )
        commands = (write_a_command, write_b_command, read_c_command)
        super().__init__(
            description='Add 2 numbers with AddNumber',
            commands=commands)

    def process_responses(self, read_responses, write_responses, resolve_future=True):
        '''
        Return the third response (from the final read command)
        Don't return any errors.
        '''
        e, result = super().process_responses(read_responses, write_responses, resolve_future=False)
        intc = result[2]
        if resolve_future:
            self.resolve_future(e, intc)
        return e, intc


class AxiAdderTest(object):
    """
    This is a test with the restriction that all the inputs must be specified before any
    out of the outputs are received.
    """

    def __init__(self):
        self.expected_intcs = []
        self.intc_futures = []

    def prepare(self, handler):
        '''
        Sends a number of 'add_numbers' commands.
        '''
        comm = AxiAdderComm(address_offset=0, handler=handler)
        n_data = 20
        max_int = pow(2, 16)-1
        logger.debug('preparing data')
        for i in range(n_data):
            inta = random.randint(0, max_int)
            intb = random.randint(0, max_int)
            self.expected_intcs.append(inta + intb)
            future = comm.add_numbers(inta, intb)
            self.intc_futures.append(future)
        # Flush the communication for simulations.
        # Ignored in FPGA.
        handler.send(comms.FakeWaitCommand(clock_cycles=10))

    def check(self):
        '''
        Check that the output of the commands matches the expected values.
        '''
        output_intcs = [f.result() for f in self.intc_futures]
        assert output_intcs == self.expected_intcs
        print('Success!!!!!!!!!!!!!!!')


def convert_to_verilog():
    working_directory = 'deleteme_axi_adder'
    core_name = 'axi_adder'
    parameters = {}
    config_filename = '/home/ben/Code/axilent/axilent/fusesoc.conf'
    filenames = fusesoc_wrapper.generate_core(
        working_directory, core_name, parameters, config_filename=config_filename, tool='vivado')
    print(filenames)
    import subprocess
    for filename in filenames:
        subprocess.call(['ghdl', '-a', filename])
    subprocess.call(['yosys', '-m', 'ghdl', '-p', 'ghdl axi_adder; write_verilog axi_adder.v'])



if __name__ == '__main__':
    convert_to_verilog()




import os
import logging
import jinja2

basedir = os.path.abspath(os.path.dirname(__file__))


def get_fusesoc_config_filename():
    template_filename = os.path.join(basedir, 'fusesoc.conf.j2')
    output_filename = os.path.join(basedir, 'fusesoc.conf')
    if not os.path.exists(output_filename):
        with open(template_filename, 'r') as f:
            template = jinja2.Template(f.read())
        content = template.render(basedir=basedir)
        with open(output_filename, 'w') as f:
            f.write(content)
    return output_filename


def setup_logging(level):
    '''
    Utility function for setting up logging.
    Configured for when slvcodec is being tested rather than used.
    '''
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    ch.setFormatter(formatter)
    # Which packages do we want to log from.
    packages = ('__main__', 'slvcodec', 'axilent')
    for package in packages:
        logger = logging.getLogger(package)
        logger.addHandler(ch)
        logger.setLevel(level)
    # Warning only packages
    packages = []
    for package in packages:
        logger = logging.getLogger(package)
        logger.addHandler(ch)
        logger.setLevel(logging.WARNING)
    logger.info('Setup logging at level {}.'.format(level))



import os
import json
import logging
import random

from slvcodec import cocotb_wrapper as cocotb
from slvcodec.cocotb_wrapper import triggers, result
from slvcodec import cocotb_dut
from slvcodec import test_utils as slvcodec_test_utils
from slvcodec import config as slvcodec_config

from axilent.examples import axi_adder
from axilent import cocotb_handler, config, test_utils


logger = logging.getLogger(__name__)

testoutput_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'test_outputs'))


fusesoc_config_filename = config.get_fusesoc_config_filename()


def test_axi_adder():
    tests = get_tests()
    vu = slvcodec_config.setup_vunit(argv=['--dont-catch-exceptions'])
    for coretest in tests:
        slvcodec_test_utils.register_coretest_with_vunit(
            vu, coretest, testoutput_dir, fusesoc_config_filename=fusesoc_config_filename)
    all_ok = vu._main(post_run=None)
    assert all_ok


@cocotb.coroutine
async def axi_adder_test(handler):
    comm = axi_adder.AxiAdderComm(address_offset=0, handler=handler)
    n_data = 100
    max_int = pow(2, 16)-1
    for i in range(n_data):
        inta = random.randint(0, max_int)
        intb = random.randint(0, max_int)
        intc = await comm.add_numbers_async(inta, intb)
        assert intc == inta + intb
    raise result.TestSuccess()


def make_handler(dut):
    axi_handler = cocotb_handler.CocotbHandler(dut.clk, m2s=dut.m2s, s2m=dut.s2m)
    axi_handler.start()
    return axi_handler


@cocotb.test()
async def axi_adder_test_top(dut):
    params_filename = os.environ['test_params_filename']
    with open(params_filename) as f:
        params = json.load(f)
    mapping = params['mapping']
    cocotb_dut.apply_mapping(dut, mapping, separator='_')
    cocotb.fork(slvcodec_test_utils.clock(dut.clk))
    dut.reset <= 0
    await triggers.RisingEdge(dut.clk)
    dut.reset <= 1
    await triggers.RisingEdge(dut.clk)
    dut.reset <= 0
    await triggers.RisingEdge(dut.clk)
    axi_handler = make_handler(dut)
    await axi_adder_test(axi_handler)


def make_coro(generics, top_params):
    async def coro(dut, resolved):
        dut.reset <= 0
        await triggers.RisingEdge(dut.clk)
        dut.reset <= 1
        await triggers.RisingEdge(dut.clk)
        dut.reset <= 0
        axi_handler = make_handler(dut)
        await axi_adder_test(axi_handler)
    return coro

def make_ftb_test(generics, top_params, resolved):
    axi_test = axi_adder.AxiAdderTest()
    test = test_utils.DictAxiTest(axi_test)
    return test

def get_tests():
    def make_test_params(resolved):
        return {}
    test = {
        'core_name': 'axi_adder',
        'entity_name': 'axi_adder',
        'generics': {'formal': False},
        'top_params': {},
        'test_module_name': 'test_axi_adder',
        'test_params': make_test_params,
        'coro': make_coro,
        'generator': make_ftb_test,
        }
    return [test]


def test_cocotb():
    test_output_directory = os.path.abspath('axi_adder_cocotb')
    tests = get_tests()
    for test in tests:
        slvcodec_test_utils.run_coretest_with_cocotb(
            test, test_output_directory,
            fusesoc_config_filename=config.get_fusesoc_config_filename(),
            generate_iteratively=False)


def test_pipes():
    test_output_directory = os.path.abspath('axi_adder_pipes')
    tests = get_tests()
    for test in tests:
        slvcodec_test_utils.run_coretest_with_pipes(
            test, test_output_directory,
            fusesoc_config_filename=config.get_fusesoc_config_filename(),
            generate_iteratively=False)


if __name__ == '__main__':
    test_cocotb()
